@@ -17,6 +17,12 @@
 #include <nettle/sha2.h>
 
 #include <string.h>
+#if FZ_WINDOWS
+#include "libfilezilla/glue/windows.hpp"
+#include <winsock.h>
+#else
+#include <arpa/inet.h>
+#endif
 
 namespace fz {
 
@@ -285,9 +291,9 @@ private:
 class hash_accumulator_hmac_sha256 final : public hash_accumulator::impl
 {
 public:
-	hash_accumulator_hmac_sha256(std::vector<uint8_t> const& key)
+	hash_accumulator_hmac_sha256(size_t keysize, uint8_t const* key)
 	{
-		nettle_hmac_sha256_set_key(&ctx_, key.size(), key.data());
+		nettle_hmac_sha256_set_key(&ctx_, keysize, key);
 	}
 
 	virtual size_t digest_size() const override { return SHA256_DIGEST_SIZE; }
@@ -312,6 +318,66 @@ private:
 	hmac_sha256_ctx ctx_;
 };
 
+class hash_accumulator_hmac_sha512 final : public hash_accumulator::impl
+{
+public:
+	hash_accumulator_hmac_sha512(size_t keysize, uint8_t const* key)
+	{
+		nettle_hmac_sha512_set_key(&ctx_, keysize, key);
+	}
+
+	virtual size_t digest_size() const override { return SHA512_DIGEST_SIZE; }
+
+	virtual void update(uint8_t const* data, size_t size) override
+	{
+		nettle_hmac_sha512_update(&ctx_, size, data);
+	}
+
+	virtual void reinit() override
+	{
+		uint8_t buf[SHA512_DIGEST_SIZE];
+		nettle_hmac_sha512_digest(&ctx_, SHA512_DIGEST_SIZE, buf);
+	}
+
+	virtual void digest(uint8_t* out) override
+	{
+		nettle_hmac_sha512_digest(&ctx_, SHA512_DIGEST_SIZE, out);
+	}
+
+private:
+	hmac_sha512_ctx ctx_;
+};
+
+class hash_accumulator_hmac_sha1 final : public hash_accumulator::impl
+{
+public:
+	hash_accumulator_hmac_sha1(size_t keysize, uint8_t const* key)
+	{
+		nettle_hmac_sha1_set_key(&ctx_, keysize, key);
+	}
+
+	virtual size_t digest_size() const override { return SHA1_DIGEST_SIZE; }
+
+	virtual void update(uint8_t const* data, size_t size) override
+	{
+		nettle_hmac_sha1_update(&ctx_, size, data);
+	}
+
+	virtual void reinit() override
+	{
+		uint8_t buf[SHA1_DIGEST_SIZE];
+		nettle_hmac_sha1_digest(&ctx_, SHA1_DIGEST_SIZE, buf);
+	}
+
+	virtual void digest(uint8_t* out) override
+	{
+		nettle_hmac_sha1_digest(&ctx_, SHA1_DIGEST_SIZE, out);
+	}
+
+private:
+	hmac_sha1_ctx ctx_;
+};
+
 hash_accumulator::hash_accumulator(hash_algorithm algorithm)
 {
 	switch (algorithm) {
@@ -333,8 +399,29 @@ hash_accumulator::hash_accumulator(hash_algorithm algorithm)
 hash_accumulator::hash_accumulator(hmac_algorithm algorithm, std::vector<uint8_t> const& key)
 {
 	switch (algorithm) {
+	case hmac_algorithm::sha512:
+		impl_ = new hash_accumulator_hmac_sha512(key.size(), key.data());
+		break;
 	case hmac_algorithm::sha256:
-		impl_ = new hash_accumulator_hmac_sha256(key);
+		impl_ = new hash_accumulator_hmac_sha256(key.size(), key.data());
+		break;
+	case hmac_algorithm::sha1:
+		impl_ = new hash_accumulator_hmac_sha1(key.size(), key.data());
+		break;
+	}
+}
+
+hash_accumulator::hash_accumulator(hmac_algorithm algorithm, std::string_view const& key)
+{
+	switch (algorithm) {
+	case hmac_algorithm::sha512:
+		impl_ = new hash_accumulator_hmac_sha512(key.size(), reinterpret_cast<uint8_t const*>(key.data()));
+		break;
+	case hmac_algorithm::sha256:
+		impl_ = new hash_accumulator_hmac_sha256(key.size(), reinterpret_cast<uint8_t const*>(key.data()));
+		break;
+	case hmac_algorithm::sha1:
+		impl_ = new hash_accumulator_hmac_sha1(key.size(), reinterpret_cast<uint8_t const*>(key.data()));
 		break;
 	}
 }
@@ -438,6 +525,18 @@ bool hash_accumulator::import_state(std::vector<std::uint8_t> const& state)
 	return ret;
 }
 
+void hash_accumulator::update_uint32_be(uint32_t v)
+{
+	v = htonl(v);
+	update(reinterpret_cast<uint8_t const*>(&v), 4);
+}
+
+void hash_accumulator::update_with_length(std::string_view const& data)
+{
+	update_uint32_be(data.size());
+	update(data);
+}
+
 namespace {
 // In C++17, require ContiguousContainer
 template<typename DataContainer>
