@@ -200,6 +200,11 @@ void client::impl::stop(bool send_done, bool keep_alive)
 		buffer_pool_->remove_waiter(*this);
 	}
 
+	if (!socket_ && waiting_for_send_) {
+		waiting_for_send_ = false;
+		remove_socket_events(this, nullptr);
+	}
+
 	if (!keep_alive) {
 		destroy_socket();
 	}
@@ -354,7 +359,9 @@ continuation client::impl::on_send()
 		}
 
 		auto root = dynamic_cast<fz::socket*>(socket_->root());
-		root->set_flags(fz::socket::flag_nodelay, true);
+		if (root) {
+			root->set_flags(fz::socket::flag_nodelay, true);
+		}
 
 		waiting_for_send_ = true;
 		return continuation::wait;
@@ -424,7 +431,9 @@ continuation client::impl::on_send()
 		// Enable Nagle's algorithm if we have a beefy body
 		if (req.body_->size() > 536) { // TCPv4 minimum required MSS
 			auto root = dynamic_cast<fz::socket*>(socket_->root());
-			root->set_flags(fz::socket::flag_nodelay, false);
+			if (root) {
+				root->set_flags(fz::socket::flag_nodelay, false);
+			}
 		}
 #if FZ_WINDOWS
 		// TCP send buffer autotuning
@@ -508,7 +517,9 @@ continuation client::impl::on_send()
 	stop_timer(buffer_tuning_timer_);
 #endif
 	auto root = dynamic_cast<fz::socket*>(socket_->root());
-	root->set_flags(fz::socket::flag_nodelay, false);
+	if (root) {
+		root->set_flags(fz::socket::flag_nodelay, false);
+	}
 
 	++send_pos_;
 	request_send_state_ = request_send_state::none;
@@ -848,6 +859,10 @@ continuation client::impl::process_complete_header()
 		return continuation::error;
 	}
 
+	if (!res.keep_alive()) {
+		wait_for_response_before_send_ = true;
+	}
+
 	auto retry = res.get_header("Retry-After");
 	if (res.code_ >= 400 && !retry.empty()) {
 		// TODO: Retry-After for redirects
@@ -1038,6 +1053,11 @@ continuation client::impl::parse_chunk_header()
 		unsigned char const* end = recv_buffer_.get() + i;
 		unsigned char* q;
 		for (q = recv_buffer_.get(); q != end && *q != ';' && *q != ' '; ++q) {
+			if (size & 0xf000000000000000ull) {
+				// Invalid size
+				logger_.log(logmsg::error, fztranslate("Malformed chunk data: %s"), fztranslate("Invalid chunk size"));
+				return continuation::error;
+			}
 			size *= 16;
 			if (*q >= '0' && *q <= '9') {
 				size += *q - '0';
