@@ -139,13 +139,13 @@ local_filesys::type local_filesys::get_file_type(native_string const& path, bool
 
 namespace {
 #ifndef FZ_WINDOWS
-local_filesys::type get_file_info_impl(int(*do_stat)(struct stat & buf, char const* path, DIR* dir, bool follow), char const* path, DIR* dir, bool &is_link, int64_t* size, datetime* modification_time, int *mode, bool follow_links)
+result get_file_info_impl(result(*do_stat)(struct stat & buf, char const* path, DIR* dir, bool follow), char const* path, DIR* dir, bool &is_link, local_filesys::type& type, int64_t* size, datetime* modification_time, int *mode, bool follow_links)
 {
 	struct stat buf{};
 	static_assert(sizeof(buf.st_size) >= 8, "The st_size member of struct stat must be 8 bytes or larger.");
 
-	int result = do_stat(buf, path, dir, false);
-	if (result) {
+	auto result = do_stat(buf, path, dir, false);
+	if (!result) {
 		is_link = false;
 		if (size) {
 			*size = -1;
@@ -156,7 +156,7 @@ local_filesys::type get_file_info_impl(int(*do_stat)(struct stat & buf, char con
 		if (modification_time) {
 			*modification_time = datetime();
 		}
-		return local_filesys::unknown;
+		return result;
 	}
 
 #ifdef S_ISLNK
@@ -175,7 +175,7 @@ local_filesys::type get_file_info_impl(int(*do_stat)(struct stat & buf, char con
 				if (modification_time) {
 					*modification_time = datetime();
 				}
-				return local_filesys::unknown;
+				return result;
 			}
 		}
 		else {
@@ -193,7 +193,8 @@ local_filesys::type get_file_info_impl(int(*do_stat)(struct stat & buf, char con
 			if (size) {
 				*size = -1;
 			}
-			return local_filesys::link;
+			type = local_filesys::link;
+			return { result::ok, 0 };
 		}
 	}
 	else
@@ -216,33 +217,67 @@ local_filesys::type get_file_info_impl(int(*do_stat)(struct stat & buf, char con
 		if (size) {
 			*size = -1;
 		}
-		return local_filesys::dir;
+		type = local_filesys::dir;
+		return { result::ok, 0 };
 	}
 
 	if (size) {
 		*size = buf.st_size;
 	}
 
-	return local_filesys::file;
+	type = local_filesys::file;
+	return { result::ok, 0 };
 }
 
-local_filesys::type get_file_info_at(char const* path, DIR* dir, bool &is_link, int64_t* size, datetime* modification_time, int *mode, bool follow)
+static result stat2result(int res)
 {
-	auto do_stat = [](struct stat & buf, char const* path, DIR * dir, bool follow)
+	if (res) {
+		int err = errno;
+
+		switch (err) {
+		case 0:
+			break;
+		case EACCES:
+		case EPERM:
+			return {result::noperm, err};
+		default:
+			return {result::other, err};
+		}
+	}
+
+	return {result::ok, 0};
+}
+
+result get_file_info_at(char const* path, DIR* dir, bool &is_link, local_filesys::type& type, int64_t* size, datetime* modification_time, int *mode, bool follow)
+{
+	auto do_stat = [](struct stat & buf, char const* path, DIR * dir, bool follow) -> result
 	{
-		return fstatat(dirfd(dir), path, &buf, follow ? 0 : AT_SYMLINK_NOFOLLOW);
+		return stat2result(fstatat(dirfd(dir), path, &buf, follow ? 0 : AT_SYMLINK_NOFOLLOW));
 	};
-	return get_file_info_impl(do_stat, path, dir, is_link, size, modification_time, mode, follow);
+	return get_file_info_impl(do_stat, path, dir, is_link, type, size, modification_time, mode, follow);
 }
 #endif
 
-local_filesys::type do_get_file_info(native_string const& path, bool& is_link, int64_t* size, datetime* modification_time, int* mode, bool follow_links)
+result do_get_file_info(native_string const& path, bool& is_link, local_filesys::type& type, int64_t* size, datetime* modification_time, int* mode, bool follow_links)
 {
 #ifdef FZ_WINDOWS
+	static auto last_result = []() -> result {
+		DWORD const err = GetLastError();
+		switch (err) {
+		case ERROR_ACCESS_DENIED:
+		case ERROR_LOGON_FAILURE:
+			return {result::noperm, err};
+		default:
+			return {result::other, err};
+		}
+	};
+
 	is_link = false;
 
 	WIN32_FILE_ATTRIBUTE_DATA data;
 	if (!GetFileAttributesEx(path.c_str(), GetFileExInfoStandard, &data)) {
+		auto result = last_result();
+
 		if (size) {
 			*size = -1;
 		}
@@ -252,7 +287,8 @@ local_filesys::type do_get_file_info(native_string const& path, bool& is_link, i
 		if (modification_time) {
 			*modification_time = datetime();
 		}
-		return local_filesys::unknown;
+
+		return result;
 	}
 
 	bool is_dir = (data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) != 0;
@@ -275,8 +311,8 @@ local_filesys::type do_get_file_info(native_string const& path, bool& is_link, i
 			if (size) {
 				*size = -1;
 			}
-
-			return local_filesys::link;
+			type = local_filesys::link;
+			return {result::ok, 0};
 		}
 
 		HANDLE hFile = is_dir ? INVALID_HANDLE_VALUE : CreateFile(path.c_str(), FILE_READ_ATTRIBUTES | FILE_READ_EA, FILE_SHARE_READ | FILE_SHARE_WRITE | FILE_SHARE_DELETE, nullptr, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, nullptr);
@@ -299,14 +335,16 @@ local_filesys::type do_get_file_info(native_string const& path, bool& is_link, i
 					if (size) {
 						*size = -1;
 					}
-					return local_filesys::dir;
+					type = local_filesys::dir;
+					return {result::ok, 0};
 				}
 
 				if (size) {
 					*size = make_int64fzT(info.nFileSizeHigh, info.nFileSizeLow);
 				}
 
-				return local_filesys::file;
+				type = local_filesys::file;
+				return {result::ok, 0};
 			}
 		}
 
@@ -319,7 +357,8 @@ local_filesys::type do_get_file_info(native_string const& path, bool& is_link, i
 		if (modification_time) {
 			*modification_time = datetime();
 		}
-		return is_dir ? local_filesys::dir : local_filesys::unknown;
+		type = is_dir ? local_filesys::dir : local_filesys::unknown;
+		return {result::ok, 0};
 	}
 
 	if (modification_time) {
@@ -337,44 +376,56 @@ local_filesys::type do_get_file_info(native_string const& path, bool& is_link, i
 		if (size) {
 			*size = -1;
 		}
-		return local_filesys::dir;
+		type = local_filesys::dir;
+		return {result::ok,0};
 	}
 	else {
 		if (size) {
 			*size = make_int64fzT(data.nFileSizeHigh, data.nFileSizeLow);
 		}
-		return local_filesys::file;
+		type = local_filesys::file;
+		return {result::ok, 0};
 	}
 #else
 	auto do_stat = [](struct stat& buf, char const* path, DIR*, bool follow)
 	{
 		if (follow) {
-			return stat(path, &buf);
-		}
-		else {
-			return lstat(path, &buf);
+			stat2result(stat(path, &buf));
 		}
+
+		return stat2result(lstat(path, &buf));
 	};
-	return get_file_info_impl(do_stat, path.c_str(), nullptr, is_link, size, modification_time, mode, follow_links);
+	return get_file_info_impl(do_stat, path.c_str(), nullptr, is_link, type, size, modification_time, mode, follow_links);
 #endif
 }
 }
 
-local_filesys::type local_filesys::get_file_info(native_string const& path, bool &is_link, int64_t* size, datetime* modification_time, int *mode, bool follow_links)
+local_filesys::type local_filesys::get_file_info(native_string const& path, bool& is_link, int64_t* size, datetime* modification_time, int *mode, bool follow_links)
+{
+	local_filesys::type type;
+
+	if (!get_file_info(path, is_link, type, size, modification_time, mode, follow_links)) {
+		type = local_filesys::unknown;
+	}
+
+	return type;
+}
+
+fz::result local_filesys::get_file_info(native_string const& path, bool& is_link, type& type, int64_t* size, datetime* modification_time, int *mode, bool follow_links)
 {
 #ifdef FZ_WINDOWS
 	if (path.size() == 6 && path[0] == '\\' && path[1] == '\\' && path[2] == '?' && path[3] == '\\' && is_drive_letter(path[4]) && path[5] == ':') {
-		return do_get_file_info(path + L"\\", is_link, size, modification_time, mode, follow_links);
+		return do_get_file_info(path + L"\\", is_link, type, size, modification_time, mode, follow_links);
 	}
 	if (path.size() == 7 && path[0] == '\\' && path[1] == '\\' && path[2] == '?' && path[3] == '\\' && is_drive_letter(path[4]) && path[5] == ':' && is_separator(path[6])) {
-		return do_get_file_info(path, is_link, size, modification_time, mode, follow_links);
+		return do_get_file_info(path, is_link, type, size, modification_time, mode, follow_links);
 	}
 #endif
 	if (path.size() > 1 && is_separator(path.back())) {
-		return do_get_file_info(path.substr(0, path.size() - 1), is_link, size, modification_time, mode, follow_links);
+		return do_get_file_info(path.substr(0, path.size() - 1), is_link, type, size, modification_time, mode, follow_links);
 	}
 
-	return do_get_file_info(path, is_link, size, modification_time, mode, follow_links);
+	return do_get_file_info(path, is_link, type, size, modification_time, mode, follow_links);
 }
 
 result local_filesys::begin_find_files(native_string path, bool dirs_only, bool query_symlink_targets)
@@ -398,6 +449,7 @@ result local_filesys::begin_find_files(native_string path, bool dirs_only, bool
 		auto const err = GetLastError();
 		switch (err) {
 		case ERROR_ACCESS_DENIED:
+		case ERROR_LOGON_FAILURE:
 			return result{result::noperm, err};
 		case ERROR_TOO_MANY_OPEN_FILES:
 			return {result::resource_limit, err};
@@ -656,7 +708,9 @@ bool local_filesys::get_next_file(native_string& name)
 #if HAVE_STRUCT_DIRENT_D_TYPE
 			if (entry->d_type == DT_LNK) {
 				bool wasLink{};
-				if (get_file_info_at(entry->d_name, dir_, wasLink, nullptr, nullptr, nullptr, query_symlink_targets_) != dir) {
+				local_filesys::type type;
+
+				if (!get_file_info_at(entry->d_name, dir_, wasLink, type, nullptr, nullptr, nullptr, query_symlink_targets_) || type != dir) {
 					continue;
 				}
 			}
@@ -666,7 +720,8 @@ bool local_filesys::get_next_file(native_string& name)
 #else
 			// Solaris doesn't have d_type
 			bool wasLink{};
-			if (get_file_info_at(entry->d_name, dir_, wasLink, nullptr, nullptr, nullptr, query_symlink_targets_) != dir) {
+			local_filesys::type type;
+			if (!get_file_info_at(entry->d_name, dir_, wasLink, type, nullptr, nullptr, nullptr, query_symlink_targets_) || type != dir) {
 				continue;
 			}
 #endif
@@ -804,7 +859,7 @@ bool local_filesys::get_next_file(native_string& name, bool &is_link, local_file
 #if HAVE_STRUCT_DIRENT_D_TYPE
 		if (dirs_only_) {
 			if (entry->d_type == DT_LNK) {
-				if (get_file_info_at(entry->d_name, dir_, is_link, size, modification_time, mode, query_symlink_targets_) != dir) {
+				if (!get_file_info_at(entry->d_name, dir_, is_link, t, size, modification_time, mode, query_symlink_targets_) || t != dir) {
 					continue;
 				}
 
@@ -818,7 +873,10 @@ bool local_filesys::get_next_file(native_string& name, bool &is_link, local_file
 		}
 #endif
 
-		t = get_file_info_at(entry->d_name, dir_, is_link, size, modification_time, mode, query_symlink_targets_);
+		if (!get_file_info_at(entry->d_name, dir_, is_link, t, size, modification_time, mode, query_symlink_targets_)) {
+			t = unknown;
+		}
+
 		if (t == unknown) { // Happens for example in case of permission denied
 #if HAVE_STRUCT_DIRENT_D_TYPE
 			t = (entry->d_type == DT_DIR) ? dir : file;
@@ -859,32 +917,65 @@ datetime local_filesys::get_modification_time(native_string const& path)
 	return mtime;
 }
 
-bool local_filesys::set_modification_time(native_string const& path, datetime const& t)
+result local_filesys::set_modification_time(native_string const& path, datetime const& t)
 {
 	if (t.empty()) {
-		return false;
+		return {result::invalid, 0};
 	}
 
 #ifdef FZ_WINDOWS
 	FILETIME ft = t.get_filetime();
 	if (!ft.dwHighDateTime) {
-		return false;
+		return {result::invalid, 0};
 	}
 
 	HANDLE h = CreateFile(path.c_str(), GENERIC_WRITE, FILE_SHARE_WRITE | FILE_SHARE_READ, nullptr, OPEN_EXISTING, 0, nullptr);
 	if (h == INVALID_HANDLE_VALUE) {
-		return false;
+		DWORD const err = GetLastError();
+		switch (err) {
+			case ERROR_FILE_NOT_FOUND:
+			case ERROR_PATH_NOT_FOUND:
+				return {result::nofile, err};
+			case ERROR_ACCESS_DENIED:
+			case ERROR_LOGON_FAILURE:
+				return {result::noperm, err};
+			default:
+				return {result::other, err};
+		}
 	}
 
 	bool ret = SetFileTime(h, nullptr, &ft, &ft) == TRUE;
+	DWORD err = GetLastError();
 	CloseHandle(h);
-	return ret;
+	if (!ret) {
+		// Dunnno how SetFileTime could have failed on a handle opened for writing
+		return {result::other, err};
+	}
+	return {};
 #else
 	struct timespec times[2]{};
 	times[0].tv_nsec = UTIME_OMIT;
 	times[1].tv_sec = t.get_time_t();
 	times[1].tv_nsec = t.get_milliseconds() * 1000000;
-	return utimensat(AT_FDCWD, path.c_str(), times, 0) == 0;
+	int res = utimensat(AT_FDCWD, path.c_str(), times, 0);
+	if (!res) {
+		return {};
+	}
+	int err = errno;
+	switch (err) {
+	case EINVAL:
+	case ENAMETOOLONG:
+		return {result::invalid, err};
+	case EPERM:
+	case EACCES:
+	case EROFS:
+	case ESRCH:
+		return {result::noperm, err};
+	case ENOENT:
+		return {result::nofile, err};
+	default:
+		return {result::other, err};
+	}
 #endif
 }
 
@@ -960,6 +1051,7 @@ result do_mkdir(native_string const& path, mkdir_permissions permissions)
 	auto const err = GetLastError();
 	switch (err) {
 	case ERROR_ACCESS_DENIED:
+	case ERROR_LOGON_FAILURE:
 		return {result::noperm, err};
 	case ERROR_DISK_FULL:
 	case ERROR_DISK_QUOTA_EXCEEDED:
@@ -1123,6 +1215,7 @@ result remove_dir(native_string const& absolute_path, bool missing_dir_is_error)
 					return {result::ok};
 				}
 			case ERROR_ACCESS_DENIED:
+			case ERROR_LOGON_FAILURE:
 				return {result::noperm, err};
 			default:
 				return {result::other, err};
@@ -1223,6 +1316,7 @@ result rename_file(native_string const& source, native_string const& dest, bool
 		case ERROR_PATH_NOT_FOUND:
 		return {result::nodir, err};
 		case ERROR_ACCESS_DENIED:
+		case ERROR_LOGON_FAILURE:
 			return {result::noperm, err};
 		case ERROR_DISK_FULL:
 		case ERROR_DISK_QUOTA_EXCEEDED:
