@@ -25,6 +25,7 @@ public:
 	void test_duplex();
 	void test_duplex_tls();
 
+	void do_test_tls_resumption(std::optional<fz::tls_ver> ver, bool server_no_ticket, bool client_no_ticket, bool use_hostname);
 	void do_test_tls_resumption(std::optional<fz::tls_ver> ver, bool server_no_ticket, bool client_no_ticket);
 	void test_tls_resumption();
 };
@@ -191,7 +192,7 @@ struct base : public fz::event_handler
 
 struct client final : public base
 {
-	client(fz::event_loop & loop, bool tls = false, std::vector<uint8_t> const& tls_session_parameters = {}, std::optional<fz::tls_ver> ver = {}, bool no_tickets = false)
+	client(fz::event_loop & loop, bool tls = false, std::vector<uint8_t> const& tls_session_parameters = {}, std::optional<fz::tls_ver> ver = {}, bool no_tickets = false, fz::native_string const& hostname = {})
 		: base(loop, tls_session_parameters, ver)
 	{
 		s_ = std::make_unique<fz::socket>(pool_, this);
@@ -206,7 +207,7 @@ struct client final : public base
 			if (no_tickets) {
 				flags |= fz::tls_client_flags::debug_no_tickets;
 			}
-			if (!tls_->client_handshake(std::vector<uint8_t>(cert.cbegin(), cert.cend()), tls_session_parameters_, {}, flags)) {
+			if (!tls_->client_handshake(std::vector<uint8_t>(cert.cbegin(), cert.cend()), tls_session_parameters_, hostname, flags)) {
 				fail(__LINE__);
 			}
 			si_ = tls_.get();
@@ -377,6 +378,12 @@ void socket_test::test_duplex_tls()
 }
 
 void socket_test::do_test_tls_resumption(std::optional<fz::tls_ver> ver, bool server_no_ticket, bool client_no_ticket)
+{
+	do_test_tls_resumption(ver, server_no_ticket, client_no_ticket, false);
+	do_test_tls_resumption(ver, server_no_ticket, client_no_ticket, true);
+}
+
+void socket_test::do_test_tls_resumption(std::optional<fz::tls_ver> ver, bool server_no_ticket, bool client_no_ticket, bool use_hostname)
 {
 	std::vector<uint8_t> server_parameters;
 	std::vector<uint8_t> client_parameters;
@@ -397,11 +404,11 @@ void socket_test::do_test_tls_resumption(std::optional<fz::tls_ver> ver, bool se
 		CPPUNIT_ASSERT(!ip.empty());
 
 		fz::event_loop client_loop;
-		client c(client_loop, true, client_parameters, ver, client_no_ticket);
+		client c(client_loop, true, client_parameters, ver, client_no_ticket, use_hostname ? fzT("localhost") : fzT(""));
 		c.handshake_only_ = true;
 
 		CPPUNIT_ASSERT(c.si_);
-		CPPUNIT_ASSERT(!c.si_->connect(ip, port));
+		CPPUNIT_ASSERT(!c.si_->connect(ip, port, fz::address_type::ipv4));
 
 		{
 			fz::scoped_lock l(c.m_);
